import numpy as np
from scipy.stats import norm
from sklearn.preprocessing import MinMaxScaler

# *******************************************************
# ****** UCB ******
# *******************************************************

def UCB(mean, std, xi):
    # Predict the model
    af = mean + xi*std
    
    return af

# *******************************************************
# ****** PI ******
# *******************************************************

def PI(mean, std, xi, x_best):

    with np.errstate(divide='warn'):
        imp = mean - x_best - xi
        Z = imp / std
        af = norm.cdf(Z)
        af[std == 0.0] = 0.0

    return af

# *******************************************************
# ****** EI ******
# *******************************************************

def EI(mean, std, xi, x_best):
    
    with np.errstate(divide='warn'):
        imp = mean - x_best - xi
        Z = imp / std
        af = imp * norm.cdf(Z) + std * norm.pdf(Z)
        af[std == 0.0] = 0.0
    
    return af

# *******************************************************
# ****** PoF ******
# *******************************************************

def PoF(x, models, engine):
    
    pof = []

    for model in models:
        if engine == 'gpflow':
            mean, std = model.predict_f(x)
            mean, var = mean.numpy(), var.numpy()
        elif engine == 'sklearn':
            mean, std = model.predict(x, return_std=True)
        elif engine == 'GPy':
            mean, std = model.predict(x)
        else:
            pass
        with np.errstate(divide='warn'):
            Z = -mean/std
            af = norm.cdf(Z)
            af[std == 0.0] = 0.0
            pof.append(af)

    return np.prod(np.array(pof), axis=0)

# *******************************************************
# ****** Prob_GPC ******
# *******************************************************

def Prob_GPC(x, model, engine):

    if engine == 'gpflow':
        Fmean, _ = model.predict_f(x)
        gpc = model.likelihood.invlink(Fmean)
        gpc = gpc.numpy()
    elif engine == 'sklearn':
        gpc = model.predict_proba(x)[:,1]
        gpc = gpc.reshape(-1, 1)
    elif engine == 'GPy':
        gpc, _ = model.predict(x)

    return gpc

# *******************************************************
# ****** AF ******
# *******************************************************

def AF(x, params, constraints_method, model, models_const, engine):

    xi, _, x_best, AF_name = params.values()

    if engine == 'gpflow':
        mean, std = model.predict_f(x)
        mean, std = mean.numpy(), std.numpy()
    elif engine == 'sklearn':
        mean, std = model.predict(x, return_std=True)
    elif engine == 'GPy':
        mean, std = model.predict(x)
    else:
        pass

    if AF_name == 'UCB':
        score = UCB(mean, std, xi)
    elif AF_name == 'PI':
        score = PI(mean, std, xi, x_best)
    elif AF_name == 'EI':
        score = EI(mean, std, xi, x_best)
    else:
        pass

    if models_const is None:
        pass
    else:
        if constraints_method == "PoF":
            score_const = PoF(x, models_const, engine)
            score_const = score_const.reshape(-1,1)
        elif constraints_method == "GPC":
            score_const = Prob_GPC(x, models_const, engine)
        else:
            pass
        scaler = MinMaxScaler()
        score_const = scaler.fit_transform(score_const)
        score = score.reshape(-1,1)
        score_const = score_const.reshape(-1,1)
        #print(score.shape, score_const.shape)
        score = score*score_const

    return score



# *******************************************************
# Import libraries
# *******************************************************

import numpy as np
import sympy as sp
from sklearn.mixture import GaussianMixture
from scipy.optimize import minimize
from scipy.stats import t
from ..utils.Aux import Flatten

# *******************************************************
# ****** Find_descriptors ******
# *******************************************************

def Find_descriptors(connected_elements, n_elements, n_jobs, dims, alpha, af_params, constraints_method, model, models_const, engine, Acq_fun):

    # *************************
    # Statistical_descriptors
    # *************************

    def Statistical_descriptors(connected_elements, n_elements, dims, alpha):

        # x_c0_rand
        def Desc_one_dims(matrix, n_elements, alpha):

            mu = [np.mean(matrix[i], axis=0).reshape(-1, 1) for i in range(n_elements)]
            sigma = []
            t_critical = []

            for i in range(n_elements):
                n = len(matrix[i])
                df = n - 1
                t_critical.append(t.ppf((1 + alpha) / 2, df).item())
                var = np.std(matrix[i].T)
                if var == 0 or len(matrix[i]) < 3:
                    sigma.append(1e-6)
                else:
                    sigma.append(var/np.sqrt(n))

            return mu, sigma, t_critical

        # x_c0_rand
        def Desc_high_dims(matrix, n_elements, dims):

            mu = [np.mean(matrix[i], axis=0).reshape(-1, 1) for i in range(n_elements)]
            Sigma = []
            Sigma_inv = []

            for i in range(n_elements):
                if (len(matrix[i]) < 3):
                    cov = np.eye(dims)
                    Sigma.append(cov)
                    Sigma_inv.append(np.linalg.inv(cov))
                else:
                    # Covariance matrix
                    cov = np.cov(matrix[i].T)
                    # Check if the covariance matrix is singular. If so, add a small value to the diagonal
                    if np.linalg.det(cov) == 0:
                        cov = cov + 1e-6*np.eye(dims)
                    Sigma.append(cov)
                    # Inverse of the covariance matrix
                    Sigma_inv.append(np.linalg.inv(cov))

            return mu, Sigma_inv

        # Main program

        if dims == 1:
            mu, Sigma_inv, t_critical = Desc_one_dims(connected_elements, n_elements, alpha)
        else:
            mu, Sigma_inv = Desc_high_dims(connected_elements, n_elements, dims)
            t_critical = None

        return mu, Sigma_inv, t_critical
    
    # *************************
    # Clusters_GMM
    # *************************    

    def Clusters_GMM(connected_elements, n_jobs, dims, af_params, constraints_method, model, models_const, engine, Acq_fun):

        x_filtred = np.concatenate(connected_elements)
        f_filtred = Acq_fun(x_filtred, af_params, constraints_method, model, models_const, engine)
        gmm = GaussianMixture(n_components=n_jobs, random_state=0).fit(np.hstack((x_filtred, f_filtred.reshape(-1,1))))
        mu = gmm.means_[:,0:dims]
        mu = [mu[i].reshape(-1,1) for i in range(n_jobs)]
        Sigma = [gmm.covariances_[i][0:dims,0:dims] for i in range(n_jobs)]
        Sigma = np.array(Sigma)
        Sigma_inv = [np.linalg.inv(Sigma[i]) for i in range(n_jobs)]

        return mu, Sigma_inv
    
    # *************************
    # Main program
    # *************************

    if n_elements < n_jobs:
        mu, Sigma_inv = Clusters_GMM(connected_elements, n_jobs, dims, af_params, constraints_method, model, models_const, engine, Acq_fun)
        if dims == 1:
            t_critical = []
            for i in range(n_elements):
                n = len(connected_elements[i])
                df = n - 1
                t_critical.append(t.ppf((1 + alpha) / 2, df).item())
        else:
            t_critical = None
    else:
        if dims == 1:
            mu, Sigma_inv, t_critical = Statistical_descriptors(connected_elements, n_elements, dims, alpha)
        else:
            mu, Sigma_inv, t_critical = Statistical_descriptors(connected_elements, n_elements, dims, alpha)

    return mu, Sigma_inv, t_critical

# *******************************************************
# ****** Find_constrains ******
# *******************************************************

def Find_constrains(x_symb, mu, Sigma_inv, chi, t_critical, n_jobs, dims):

    # *************************
    # CI_one_dims
    # *************************    

    def CI_one_dims(n_jobs, mu, sigma, t_critical):

        CI = []

        for i in range(n_jobs):
            margin_of_error = t_critical[i] * sigma[i]
            CI.append([mu[i].item() - margin_of_error, mu[i].item() + margin_of_error])
        
        return CI
    
    # *************************
    # CI_high_dims
    # *************************
    
    def CI_high_dims(x_symb, n_jobs, mu, Sigma_inv, chi):

        CI = []

        for i in range(n_jobs):
            y = (x_symb-mu[i])
            dist = y.T*Sigma_inv[i]*y
            CI.append(sp.LessThan(dist[0], chi))
            
        return CI
    
    # *************************
    # Lambify_CI
    # *************************
    
    def Lambify_CI(x_symb, CI, n_jobs, dims):
        
        # Lamba_fun
        def Lamba_fun(fun):

            lam = lambda x: fun(*x)
            
            return lam
        
        if dims == 1:
            CI_lambda = CI
        else:
            const_lamb = [sp.lambdify(x_symb, (CI[i].rhs - CI[i].lhs), 'numpy') for i in range(n_jobs)]
            CI_lambda = [Lamba_fun(const_lamb[i]) for i in range(n_jobs)]
        
        return CI_lambda
    
    # *************************
    # Main program
    # *************************

    if dims == 1:                           
        CI = CI_one_dims(n_jobs, mu, Sigma_inv, t_critical)
    else:
        CI = CI_high_dims(x_symb, n_jobs, mu, Sigma_inv, chi)

    return Lambify_CI(x_symb, CI, n_jobs, dims)

# *******************************************************
# ****** Querry ******
# *******************************************************

def Querry(n_jobs, mu, CI_lambda, bnds, dims, af_params, constraints_method, sense, model, models_const, engine, Acq_fun):

    # Acquisition function for the optimization
    def Acquisition_function_opt(x, af_params, sense, model, models_const, engine):

        if sense == "maximize":
            af_opt = -Acq_fun(x.reshape(1,-1), af_params, constraints_method, model, models_const, engine)
        elif sense == "minimize":
            af_opt = Acq_fun(x.reshape(1,-1), af_params, constraints_method, model, models_const, engine)
        
        return af_opt
    
    # *************************
    # Main program
    # *************************
    
    x_opt = []

    for i in range(n_jobs):
        # Define the optimization problem
        if dims == 1:
            bnds = tuple(map(tuple, CI_lambda))
            res = minimize(Acquisition_function_opt, x0=mu[i].flatten(), args=(af_params, sense, model, models_const, engine), method='BFGS', bounds=(bnds[i],))
        else:
            res = minimize(Acquisition_function_opt, x0=mu[i].flatten(), args=(af_params, sense, model, models_const, engine), method='SLSQP', bounds=bnds, constraints={'type': 'ineq', 'fun': CI_lambda[i]})
        x_opt.append(res.x)

    return x_opt


from .Acq_fun import UCB, PI, EI, PoF, Prob_GPC, AF
from .Aux import Errors, Flatten, Eval_fun, Eval_const, Imputation, Best_values, Check_if_improvement, Regret, Print_header, Print_results, Create_results
from .Initialize import Space, Problem_type, Get_constraints, Get_n_p_design, U_Generator, Get_x_and_z, Get_constraints, Get_search_space_params, Get_x_mesh, Bounds, Get_n_jobs, Get_kernel, Iter_params, AF_params
from .Models import Train_model, Train_models_constraints, Kernel_discovery
from .Preprocessing_data import U_scaling, U_inverse_scaling, Trans_data_to_pandas, Reduce, Inverse, Train_reducer, Train_inverter, Find_reducer, Find_inverter, Red_bounds, X_new_scaling
from .Querry_points import Find_descriptors, Find_constrains, Querry
from .Superlevel_set_filtration import Sl_sf
from .Update import Up_search_space_params, Up_mesh


# *******************************************************
# ****** Import libraries ******
# *******************************************************

import numpy as np
from itertools import product, permutations
from operator import itemgetter
from ..utils.Aux import Flatten

# *******************************************************
# ****** Sl_sf ******
# *******************************************************

def Sl_sf(mesh, p_mesh, dims, jobs, q0, af_params, constraints_method, model, models_const, engine, sense, Acq_fun):

    """ 
    Find connected component of the matrix of points x in the domain adove the level set
    """

    # *************************
    # Filtration
    # *************************

    def Filtration(x, score, p_x, dims, level, sense):
        if sense == "maximize":
            data_matrix = np.array([score > level])[0]
        elif sense == "minimize":
            data_matrix = np.array([score < level])[0]
        # Reshape the data
        data_matrix = data_matrix.reshape(*[p_x]*dims)

        return data_matrix.astype(int), x.reshape(*[p_x]*dims, dims)
    
    # *************************
    # Find_connected_elements_dfs
    # *************************

    def Find_connected_elements_dfs(nodes, matrix):
        # Number of dimensions
        D = [nodes.shape[i] for i in range(dims)]
        D0 = D[0]
        # Define the neighbors
        s_neg = [-1]
        s_pos = [1]
        for i in range(dims-1):
            s_neg.append(0)
            s_pos.append(0)
        neigh = tuple(set(permutations(s_neg, dims)) | set(permutations(s_pos, dims)))
        
        # dfs_connected
        def dfs_connected(element, value):
            #
            elem_tuple = tuple(element)
            # Check if the current cell is outside the matrix or has a different value
            if [False for i in range(len(element)) if element[i] < 0] or [False for i in range(len(element)) if element[i] >= D[i]] or nodes[elem_tuple] != value:
                return []
            
            # Mark the cell as visited
            nodes[elem_tuple] = -1
            # Initialize the connected component with the current cell
            nodes_conected = [elem_tuple]
            # Check the neighbors
            for dr in neigh:
                nodes_conected.extend(dfs_connected([element[i] + dr[i] for i in range(len(element))], value))
            
            return nodes_conected
        
        # main program
        list_connected_components = []
        d = product(range(D0), repeat=dims)
        # cycle
        for i in d:
            if nodes[i] == 1:
                nodes_conected = dfs_connected(i, nodes[i])
                list_connected_components.append(nodes_conected)
        
        connected_elements = [np.array([matrix[tuple(list_connected_components[k][i])] for i in range(len(list_connected_components[k]))]) for k in range(len(list_connected_components))]
        
        return connected_elements
    
    # *************************
    # List_of_elements
    # *************************

    def List_of_elements(x, score, p_x, n_elements, dims, q, sense):
        #
        connected_elements = []
        level = np.percentile(np.concatenate(score), q)
        if level == 0:
            level = np.mean(np.concatenate(score))
        #
        for j in range(n_elements):
            x_flitred = Filtration(x[j], score[j], p_x[j], dims, level, sense)
            connected_elements.append(Find_connected_elements_dfs(x_flitred[0], x_flitred[1])) 

        connected_elements = [x for x in connected_elements if x != []]
        connected_elements = Flatten(connected_elements)
        n_connected_elements = len(connected_elements)

        return connected_elements, n_connected_elements
    
    # *************************
    # Reduce_num_elements
    # *************************
    
    def Reduce_num_elements(x, mesh, score, n_elements, jobs, sense):

        def Replace_val(value, arr_1, arr_2):

            return arr_2[(arr_1 == value).all(axis=1)].reshape(-1)

        score_all = np.concatenate(score)
        mesh_all = np.concatenate(mesh)
        scores_lst = []
        scores_mean = []

        for i in range(n_elements):
            try:
                scores_lst.append(np.array([Replace_val(x[i][j], mesh_all, score_all) for j in range(len(x[i]))]))
            except:
                return [Replace_val(x[i][j], mesh_all, score_all) for j in range(len(x[i]))]

        for i in range(n_elements):
            if len(scores_lst[i].shape) == 1:
                scores_mean.append(np.mean([np.mean(scores_lst[i][j]) for j in range(len(scores_lst[i]))]))
            else:
                scores_mean.append(np.mean(scores_lst[i]))
                
        if sense == "maximize":
            sorted = np.sort(scores_mean)[::-1][:jobs]
        elif sense == "minimize":
            sorted = np.sort(scores_mean)[:jobs]
        l = np.array([np.where(scores_mean == sorted[i]) for i in range(len(sorted))]).reshape(-1)
        #print(l)
        try:
            l = Flatten(l)
            l = list(set(l))
        except:
            l = l
        connected_elements = [x[i] for i in l]
        n_connected_elements = len(connected_elements)

        return connected_elements, n_connected_elements
    
    # *************************
    #  Main
    # *************************

    score = [Acq_fun(mesh[i], af_params, constraints_method, model, models_const, engine) for i in range(len(mesh))]
    n_elements = len(mesh)
    # 
    if sense == "maximize":
        connected_elements, n_connected_elements = List_of_elements(mesh, score, p_mesh, n_elements, dims, q0, sense)
    elif sense == "minimize":
        connected_elements, n_connected_elements = List_of_elements(mesh, score, p_mesh, n_elements, dims, 100-q0, sense)
    else:
        pass
    #  
    if n_connected_elements > jobs:
        try:
            connected_elements, n_connected_elements = Reduce_num_elements(connected_elements, mesh, score, n_connected_elements, jobs, sense)
        except:
            b = np.random.choice(n_connected_elements, jobs, replace=False)
            connected_elements = itemgetter(*b)(connected_elements)
            n_connected_elements = jobs
        

    return connected_elements, n_connected_elements


# *******************************************************
# ****** Import libraries ******
# *******************************************************

import numpy as np
import multiprocess as mp
from sklearn.impute import SimpleImputer
from types import SimpleNamespace

# *******************************************************
# ****** Errors ******
# *******************************************************

def Errors(d_nc, d_nd, design_type, surrogate, engine, constraints_method, AF_name):

    valid_var_type = ["continuous", "integer", "categorical"]
    valid_design_type = ["random", "LHS", "Sobol", "Halton"]
    valid_surrogate_name = ["GP", "SGP"]
    valid_engines = ["gpflow", "sklearn", "GPy"]
    valid_constraints_method = ["PoF", "GPC"]
    valid_af_names = ['UCB', 'EI', 'PI']

    if d_nc <= 0 and d_nd <= 0:
        raise ValueError("Not valid variable type, valid names are:", *valid_var_type)

    if design_type not in valid_design_type:
        raise ValueError("Not valid design type, valid names are:", *valid_design_type)
    
    if surrogate not in valid_surrogate_name:
        raise ValueError("Not valid acquisition function name, valid names are:", *valid_surrogate_name)
    
    if engine not in valid_engines:
        raise ValueError("Not valid engine name, valid names are:", *valid_engines)
    
    if constraints_method not in valid_constraints_method:
        raise ValueError("Not valid method for constraints, valid names are:", *valid_constraints_method)

    if AF_name not in valid_af_names:
        raise ValueError("Not valid acquisition function name, valid names are:", *valid_af_names)
    
    return None

# *******************************************************
# ****** Best_values ******
# *******************************************************

def Flatten(l):

    return [item for sublist in l for item in sublist]

# *******************************************************
# ****** Eval_fun ******
# *******************************************************

def Eval_fun(x, jobs, function):

    if jobs == 1:
        x_new = np.array(x).reshape(1,-1)
        f_new = function(x_new)
    else:
        x_new = [x[i].reshape(1,-1) for i in range(jobs)]
        with mp.Pool(jobs) as pool:
            f_new = pool.map(function, x_new)
        
    return f_new

# *******************************************************
# ****** Eval_constrains ******
# *******************************************************

def Eval_const(x, const, n_const, const_method):

    if const == None: 
        y = None
        g = None
    else:
        if const_method == "PoF":
            data = [eval(const[i], None, {"x": x}) for i in range(n_const)]
        elif const_method == "GPC":
            data = [eval(const[i]['constraint'], None, {"x": x}) for i in range(n_const)]
        else:
            pass

        g = np.array(data).T

        if const_method == "PoF":
            y = g <= 0
            y = (y == True).all(axis=1)
        elif const_method == "GPC":
            y = (g == True).all(axis=1)
            g = y.astype(int).reshape(-1,1)
        else:
            pass

    return g, y

# *******************************************************
# ****** Imputation ******
# *******************************************************

def Imputation(g, const, const_method):

    if const == None: 
        g_new = None
    else:
        if const_method == "PoF":
            if np.isnan(g).any() == True:
                imputer = SimpleImputer(missing_values=np.nan, strategy='mean')
                g_new = imputer.fit_transform(g)
            else:
                g_new = g
        elif const_method == "GPC":
            g_new = g
        else:
            pass

    return g_new

# *******************************************************
# ****** Best_values ******
# *******************************************************

def Best_values(x, f, y, sense):

    if y is None:
        pass
    else:
        f = f[y]
        x = x[y]

    if sense == "maximize":
        ix_best = np.argmax(f)
        f_best = np.max(f)
    elif sense == "minimize":
        ix_best = np.argmin(f)
        f_best = np.min(f)
    x_best = x[ix_best]

    return x_best, f_best

# *******************************************************
# ****** Check_if_improvement ******
# *******************************************************

def Check_if_improvement(f_new, f_best, jobs, sense):

    if jobs == 1:
        f_best_new = f_new
    else:
        if sense == "maximize":
            f_best_new = np.max(f_new)
        elif sense == "minimize":
            f_best_new = np.min(f_new)
        else:
            pass
    
    flag = 0
    
    if sense == "maximize":
        if f_best_new > f_best:
            flag = 1
    elif sense == "minimize":
        if f_best_new < f_best:
            flag = 1
    else:
        pass

    return flag

# *******************************************************
# ****** Regret ******
# *******************************************************

def Regret(f_true, x, n_elements, model, engine):

    # Return an average of the reward
    if engine == 'gpflow':
        f_pred, _ = model.predict_f(x)
    elif engine == 'sklearn':
        f_pred, _ = model.predict(x, return_std=True)
    elif engine == 'GPy':
        f_pred, _ = model.predict(x)
    else:
        pass

    rt = [(f_true[i] - f_pred[i]) for i in range(n_elements)]
    rt = sum(rt)

    return rt/n_elements

# *******************************************************
# ****** Print_header ******
# *******************************************************

def Print_header(names, x_symb_names, dims):

    if names is None:
        header = 'ite  ' +  '  f      ' + str(x_symb_names[0])
    else:
        header = 'ite  ' +  '  f      ' + str(names[0])
    for i in range(1, dims):
        if names is None:
            header += '      ' + str(x_symb_names[i])
        else:
            header += '      ' + str(names[i])
    
    return header

# *******************************************************
# ****** Print_results ******
# *******************************************************

def Print_results(x, f, cat_val):

    if cat_val == 0:
        x = x.reshape(-1)
        x_print = ["%.5f" % value if 1e-3 < abs(value) < 1e3 else "%0.1e" % value for value in x]
    else:
        #x = x[0]
        x_print = [] 
        for value in x:
            if type(value) == str:
                x_print.append(value)
            elif type(value) == float:
                if 1e-3 < abs(value) < 1e3:
                    x_print.append("%.5f" % value)
                else:
                    x_print.append("%0.1e" % value)
    
    f_print = "%.5f" % f if 1e-3 < abs(f) < 1e3 else "%0.1e" % f
        
    return x_print, f_print

# *******************************************************
# ****** Create_results ******
# *******************************************************

def Create_results(x_best, f_best, x, f, x_l, x_u, dims, max_iter, points, design, af_params, constraints_method, rt, models_const, model):

    res = {'x_best': x_best, 'f_best': f_best, 
           'x_init': x[0:points], 'f_init': f[0:points], 
           'x_iters': x[points:-1], 'f_iters': f[points:-1],
           'x_l': x_l, 'x_u': x_u, 
           'dims': dims,
           'iters': max_iter, 
           'initial_design': design,
           'initial_points': points,
           'xi': af_params['xi'], 
           'acquisition_function': af_params['AF_name'],
           'regret': rt,
           'constraint_method': constraints_method,
           'models_constraints': models_const,
           'model': model}
    
    return SimpleNamespace(**res)


# *******************************************************
# ****** Import libraries ******
# *******************************************************

import numpy as np
import pandas as pd
import properscoring as ps
from sklearn.model_selection import train_test_split
from sklearn.gaussian_process import kernels as kernels_sklear
from sklearn.gaussian_process import GaussianProcessRegressor as GPR_sklearn
from sklearn.gaussian_process import GaussianProcessClassifier as GPC_sklearn
from gpflow  import likelihoods
from gpflow import kernels as kernels_gpflow
from gpflow.models import GPR as GPR_gpflow
from gpflow.models import SGPR as SGPR_gpflow
from gpflow.models import VGP as GPC_gpflow
from gpflow.optimizers import Scipy as optimizer_gpflow

try: 
    from GPy import kern as kernels_gpy
    from GPy.models import GPRegression as GPR_gpy
    from GPy.models import SparseGPRegression as SGPR_gpy
    from GPy.models import GPClassification as GPC_gpy
except:
    pass

# *******************************************************
# ****** Train_model ******
# *******************************************************

def Train_model(x, f, kernel, surrogate, n_restarts, engine):

    def Select_model(x, f, kernel, surrogate, engine):

        if surrogate == "GP":
            if engine == 'gpflow':
                model = GPR_gpflow((x, f), kernel=kernel,)
            elif engine == 'sklearn':
                model = GPR_sklearn(kernel=kernel, n_restarts_optimizer=n_restarts)
            elif engine == 'GPy':
                model = GPR_gpy(x, f, kernel)
            else:
                pass
        elif surrogate == "SGP":
            if engine == 'gpflow':
                rng = np.random.default_rng(1234)
                n_inducing = 10
                inducing_variable = rng.choice(x, size=n_inducing, replace=False)
                model = SGPR_gpflow((x, f), kernel=kernel, inducing_variable=inducing_variable)
            elif engine == 'sklearn':
                raise ValueError("sklearn does not implement Sparse Gaussian Process")
            elif engine == 'GPy':
                model = SGPR_gpy(x, f, kernel)
            else:
                pass
        else:
            pass
        
        return model
    
    if x.dtype == object:
        x = np.asarray(x).astype(np.float64)
    
    model = Select_model(x, f, kernel, surrogate, engine)

    if engine == 'gpflow':
        opt = optimizer_gpflow()
        opt.minimize(model.training_loss, model.trainable_variables)
    elif engine == 'sklearn':
        model.fit(x, f)
    elif engine == 'GPy':
        model.optimize(optimizer='lbfgsb', max_iters=1000, messages=False)
        model.optimize_restarts(num_restarts=n_restarts, verbose=False)
    else:
        pass

    return model

# *******************************************************
# ****** Train_models_const ******
# *******************************************************

def Train_models_constraints(x, g, constraints, n_constraints, constraints_method, engine):

    
    def Train_PoF(x, g, n_const, engine):
        
        models = []
        for i in range(n_const):
            if engine == 'gpflow':
                model = GPR_gpflow((x, g[:,i].reshape(-1,1)), kernel=kernels_gpflow.SquaredExponential(),)
                opt = optimizer_gpflow()
                opt.minimize(model.training_loss, model.trainable_variables)
            elif engine == 'sklearn':
                model = GPR_sklearn().fit(x, g[:,i].reshape(-1,1))
            elif engine == 'GPy':
                model = GPR_gpy(x, g[:,i].reshape(-1,1))
                model.optimize()
                models.append(model)
            else: 
                pass

        return models
    
    def Train_GPC(x, g, engine):

        if engine == 'gpflow':
            model = GPC_gpflow((x, g),kernel=kernels_gpflow.RBF(),likelihood=likelihoods.Bernoulli(),)
            opt = optimizer_gpflow()            
            opt.minimize(model.training_loss, model.trainable_variables)
        elif engine == 'sklearn':
            model = GPC_sklearn().fit(x, g)
        elif engine == 'GPy':
            model = GPC_gpy(x, g)
            model.optimize()
        else:
            pass

        return model

    if constraints == None: 
        models_const = None
    else:
        if constraints_method == "PoF":
            models_const = Train_PoF(x, g, n_constraints, engine)
        elif constraints_method == "GPC":
            models_const = Train_GPC(x, g, engine)
        else:
            pass

    return models_const

# *******************************************************
# ****** Kernel_discovery ******
# *******************************************************

def Kernel_discovery(x, f, dims, surrogate, beta, evals, engine):

    def Search(x_train, x_test, f_train, f_test, n, kernels, surrogate, engine):

        BIC, CRPS, models = [], [], {}

        if x_train.dtype == object:
            x_train = np.asarray(x_train).astype(np.float64)
            x_test = np.asarray(x_test).astype(np.float64)
        
        for name, kernel in kernels.items():
            if surrogate == "GP":
                if engine == 'gpflow':
                    model = GPR_gpflow((x_train, f_train), kernel=kernel,)
                elif engine == 'sklearn':
                    model = GPR_sklearn(kernel=kernel)
                elif engine == 'GPy':
                    model = GPR_gpy(x_train, f_train.reshape(-1,1), kernel)
                else:
                    pass
            elif surrogate == "SGP":
                if engine == 'gpflow':
                    rng = np.random.default_rng(1234)
                    n_inducing = 10
                    inducing_variable = rng.choice(x_train, size=n_inducing, replace=False)
                    model = SGPR_gpflow((x_train, f_train), kernel=kernel, inducing_variable=inducing_variable)
                elif engine == 'sklearn':
                    raise ValueError("sklearn does not implement Sparse Gaussian Process")
                elif engine == 'GPy':
                    model = SGPR_gpy(x_train, f_train, kernel)
                else:
                    pass
            else:
                pass

            if engine == 'gpflow':
                opt = optimizer_gpflow()
                opt.minimize(model.training_loss, model.trainable_variables)
                ll, k = model.maximum_log_likelihood_objective().numpy(), sum(np.prod(v.shape) for v in model.trainable_variables)
                mean, var = model.predict_f(x_test)
                mean, var = mean.numpy(), var.numpy()
            elif engine == 'sklearn':
                model.fit(x_train, f_train)
                ll, k = model.log_marginal_likelihood_value_, len(model.kernel_.theta)
                mean, var = model.predict(x_test, return_std=True)
            elif engine == 'GPy':
                model.optimize()
                ll, k = model.log_likelihood(), model._size_transformed()
                mean, var = model.predict(x_test)
            else:
                pass
            
            BIC.append((-2*ll + k*np.log(n)).item())
            CRPS.append(ps.crps_gaussian(f_test, mean, var).mean())
            models[name] = model
        
        return BIC, CRPS, models

    def Get_best_model(BIC, CRPS, models, beta, eval):

        normalized_bics = (BIC - np.min(BIC)) / (np.max(BIC) - np.min(BIC))
        normalized_crpss = (CRPS - np.min(CRPS)) / (np.max(CRPS) - np.min(CRPS))

        # Combine scores (example: equal weight)
        
        scores = beta * normalized_bics + (1-beta) * normalized_crpss
        #df_BIC = pd.DataFrame(np.vstack((BIC, normalized_bics)), columns=models.keys())
        #df_CRPS = pd.DataFrame(np.vstack((CRPS, normalized_crpss)), columns=models.keys())
        #df_scores = pd.DataFrame(np.array(scores).reshape(1,-1), columns=models.keys())
        #df_BIC.to_csv('/Users/javiermorlet/Codes/Project_3/plots/Results/Example_1/Fig_1_3/df_BIC' + str(eval) + '.csv')
        #df_CRPS.to_csv('/Users/javiermorlet/Codes/Project_3/plots/Results/Example_1/Fig_1_3/df_CRPS' + str(eval) + '.csv')
        #df_scores.to_csv('/Users/javiermorlet/Codes/Project_3/plots/Results/Example_1/Fig_1_3/df_scores' + str(eval) + '.csv')
        models_ordered = [x for _, x in sorted(zip(scores, models.keys()))]
        
        return {models_ordered[0]: models[models_ordered[0]]}
    
    x_train, x_test, f_train, f_test = train_test_split(x, f, test_size=0.2)
    n = x_train.shape[0]
    # Base kernels: SE, Periodic, Linear, RatQuad
    if engine == 'gpflow':
        kernels_ = {"linear": kernels_gpflow.Linear(),
                "RBF": kernels_gpflow.SquaredExponential(),
                "Mater_52": kernels_gpflow.Matern52(),
                "Periodic": kernels_gpflow.Periodic(kernels_gpflow.SquaredExponential())
                }
    elif engine == 'sklearn':
        kernels_ = {"linear": kernels_sklear.DotProduct(),
                "RBF": kernels_sklear.RBF(),
                "Mater_52": kernels_sklear.Matern(nu=2.5),
                "Rational_Quadratic": kernels_sklear.RationalQuadratic()  
                }
    elif engine == 'GPy':
        kernels_ = {"linear": kernels_gpy.Linear(input_dim=dims),
                "RBF": kernels_gpy.RBF(input_dim=dims, variance=1.0, lengthscale=1.0),
                "Mater_52": kernels_gpy.Matern52(input_dim=dims, variance=1.0, lengthscale=1.0),
                "Periodic": kernels_gpy.StdPeriodic(input_dim=dims, variance=1.0, lengthscale=1.0, period=1.0)
                }
    else:
        pass

    for i in range(evals):
        base_kernels = kernels_.copy()
        BIC, CRPS, models = Search(x_train, x_test, f_train, f_test, n, base_kernels, surrogate, engine)
        best_model = Get_best_model(BIC, CRPS, models, beta, i)
        base_model = list(best_model.values())[0]
        base_model_name = list(best_model.keys())[0]
        if engine == 'gpflow':
            base_model_kern = base_model.kernel
        elif engine == 'sklearn':
            base_model_kern = base_model.kernel_
        elif engine == 'GPy':
            base_model_kern = base_model.kern
        else:
            pass
        if i == evals: break
        kernels_ = {}
        for name, kernel in base_kernels.items():
            if engine == 'gpflow':
                kernels_[base_model_name + "+" + name] = base_model_kern + kernel
                kernels_[base_model_name + "*" + name] = base_model_kern * kernel
            elif engine == 'sklearn':
                kernels_[base_model_name + "+" + name] = kernels_sklear.Sum(base_model_kern, kernel)
                kernels_[base_model_name + "*" + name] = kernels_sklear.Product(base_model_kern, kernel)
            elif engine == 'GPy':
                kernels_[base_model_name + "+" + name] = kernels_gpy.Add([base_model_kern, kernel])
                kernels_[base_model_name + "*" + name] = kernels_gpy.Prod([base_model_kern, kernel])
            else:
                pass

    return base_model_kern


"pyBOWIE (python Bayesian Optimization WIth supErlevel set filtration)"

import numpy as np
import warnings
from ..utils.Initialize import Space, Problem_type, Get_x_and_z, Get_constraints, Get_search_space_params, Get_x_mesh, Bounds, Get_n_jobs, Get_kernel, Iter_params, AF_params
from ..utils.Aux import Errors, Best_values, Check_if_improvement, Eval_fun, Eval_const, Imputation, Regret, Print_header, Print_results, Create_results
from ..utils.Preprocessing_data import Red_bounds, X_new_scaling
from ..utils.Models import Train_model, Train_models_constraints
from ..utils.Acq_fun import AF
from ..utils.Superlevel_set_filtration import Sl_sf
from ..utils.Querry_points import Find_constrains, Find_descriptors, Querry
from ..utils.Update import Up_search_space_params, Up_mesh

class BO():

    def __init__(self, function, domain, sense,
                 surrogate = "GP",
                 engine = 'gpflow',
                 acquisition_function = "UCB",
                 xi_0 = 2,
                 xi_f = 0.1,
                 xi_decay = "yes",
                 kernel = None,
                 kern_discovery = "yes",
                 kern_discovery_evals = 2,
                 x_0 = None, 
                 f_0 = None,
                 design = "LHS",
                 n_p_design = None,
                 n_jobs = None,
                 n_restarts = 5,
                 max_iter = 100, 
                 constraints = None,
                 constraints_method = "PoF",
                 reducer = None,
                 n_components = None,
                 inverter_transform = "no",
                 alpha = 0.95,
                 beta = 0.5,
                 c1_param = 50,
                 c2_param = 10,
                 verbose = 0,
                 ):

        self.function = function
        self.domain = domain
        self.sense = sense
        self.surrogate = surrogate
        self.engine = engine
        self.acquisition_function = acquisition_function
        self.xi_0 = xi_0
        self.xi_f = xi_f
        self.xi_decay = xi_decay
        self.kernel = kernel
        self.kern_discovery = kern_discovery
        self.kern_discovery_evals = kern_discovery_evals
        self.x_0 = x_0
        self.f_0 = f_0
        self.design = design
        self.n_p_design = n_p_design
        self.n_jobs = n_jobs
        self.n_restarts = n_restarts
        self.max_iter = max_iter
        self.constraints = constraints
        self.constraints_method = constraints_method
        self.reducer = reducer
        self.n_components = n_components
        self.inverter_transform = inverter_transform
        self.alpha = alpha
        self.beta = beta
        self.c1_param = c1_param
        self.c2_param = c2_param
        self.verbose = verbose

    def optimize(self):

        # *************** Main program ********************
        # Ignore warnings
        warnings.filterwarnings("ignore")
        #
        dims, d_nc, d_ni, d_nq, d_nd, x_l, x_u, int_val, cat_val, names = Space(self.domain)
        #
        Errors(d_nc, d_nd, self.design, self.surrogate, self.engine, self.constraints_method, self.acquisition_function)
        #
        problem_type = Problem_type(d_nc, d_ni, d_nq)
        # Get initial points of evaluation
        x, f, n_p_design_ = Get_x_and_z(self.function, self.x_0, self.f_0, dims, d_nc, d_ni, d_nq, x_l, x_u, int_val, cat_val, self.c1_param, problem_type, self.design, self.n_p_design)
        # Get constraints
        constraints_, n_constraints_ = Get_constraints(self.constraints, self.constraints_method)
        # Evaluate constraints
        g, y = Eval_const(x, constraints_, n_constraints_, self.constraints_method)
        g = Imputation(g, constraints_, self.constraints_method)
        # Perform data preprocessing, and dimensionality reduction and reduce if required
        x_tau, reducer_name, reducer_trained, inverter, dims_tau = Get_search_space_params(x, dims, d_nc, d_ni, d_nq, d_nd, x_l, x_u, int_val, cat_val, n_p_design_, problem_type, self.reducer, self.inverter_transform)
        # Reduce bounds
        x_l_tau, x_u_tau = Red_bounds(x_l, x_u, int_val, cat_val, d_nc, d_ni, d_nq, dims, dims_tau, problem_type, reducer_trained)
        # Get mesh
        x_mesh, n_p_mesh, n_p_mesh_0 = Get_x_mesh(x_l_tau, x_u_tau, dims_tau, self.c2_param)
        # Tuple of the bounds
        bounds = Bounds(x_l_tau, x_u_tau, dims_tau)
        # Spawning processes
        n_jobs_ = Get_n_jobs(self.n_jobs)
        # Select covariance
        kernel_ = Get_kernel(x_tau, f, dims_tau, self.surrogate, self.beta, self.kernel, self.kern_discovery, self.kern_discovery_evals, self.engine)
        # Initialize iteration parameters
        x_symb, x_symb_names, flag, q, q_inc, delta_q, chi, update_param = Iter_params(dims, dims_tau, self.max_iter, self.alpha)
        # Initialize AF parameters
        af_params = AF_params(f, self.xi_0, self.xi_f, self.xi_decay, self.max_iter, self.acquisition_function, self.sense)
        # Get best values
        x_best, f_best = Best_values(x, f, y, self.sense)
        af_params["f_best"] = f_best
        rt = []
        # Print results if vervose is active
        if self.verbose == 1:
            print(Print_header(names, x_symb_names, dims))
            x_print, f_print = Print_results(x_best, f_best, cat_val)
            print(0, ' ', f_print, *x_print)
        # *****
        # Cycle
        # *****
        for ite in range(1, self.max_iter+1):
            # Train surrogate model
            model = Train_model(x_tau, f, kernel_, self.surrogate, self.n_restarts, self.engine)
            # Train surrogate model for constraints
            models_constraints = Train_models_constraints(x_tau, g, constraints_, n_constraints_, self.constraints_method, self.engine)
            # Find conected elements
            connected_elements, n_connected_elements = Sl_sf(x_mesh, n_p_mesh, dims_tau, n_jobs_, q, af_params, self.constraints_method, model, models_constraints, self.engine, self.sense, AF)
            # Find CI bounds
            try:
                conc = np.concatenate(connected_elements)
            except:
                conc = np.hstack(connected_elements)
            try:
                if n_connected_elements == 0 or len(conc) < n_jobs_:
                    mu, Sigma_inv, t_critical = Find_descriptors(x_mesh, len(x_mesh), n_jobs_, dims_tau, self.alpha, af_params, self.constraints_method, model, models_constraints, self.engine, AF)
                else:
                    mu, Sigma_inv, t_critical = Find_descriptors(connected_elements, n_connected_elements, n_jobs_, dims_tau, self.alpha, af_params, self.constraints_method, model, models_constraints, self.engine, AF)
            except:
                print(n_connected_elements, conc)
            # Lambdify CI bounds
            CI_lambda = Find_constrains(x_symb, mu, Sigma_inv, chi, t_critical, n_jobs_, dims_tau)
            # Find new point(s)
            x_new_tau = Querry(n_jobs_, mu, CI_lambda, bounds, dims_tau, af_params, self.constraints_method, self.sense, model, models_constraints, self.engine, AF)
            # Scale new point(s)
            x_new = X_new_scaling(x_new_tau, d_nc, d_ni, d_nq, x_l, x_u, int_val, cat_val, n_jobs_, reducer_name, self.inverter_transform, inverter, problem_type)
            # Evaluate objective function
            f_new = Eval_fun(x_new, n_jobs_, self.function)
            # Evaluate constraints
            g_new, y_new = Eval_const(x_new, constraints_, n_constraints_, self.constraints_method)
            # Update train data
            x, x_tau, f = np.vstack((x, x_new)), np.vstack((x_tau, x_new_tau)), np.vstack((f, f_new))
            if g is None:
                pass
            else:
                g = np.vstack((g, g_new))
                y = np.hstack((y, y_new))
                g = Imputation(g, constraints_, self.constraints_method)
            # Update reducer, inverter, inducing points
            if (ite % update_param) == 0:
                x_tau, connected_elements, reducer_trained, inverter = Up_search_space_params(x, x_tau, connected_elements, n_connected_elements, d_nc, dims, problem_type, dims_tau, reducer_name, inverter, self.inverter_transform)
            # Update x_mesh if the new point improves objetive function
            flag = Check_if_improvement(f_new, f_best, n_jobs_, self.sense)
            if flag == 1:
                try:
                    x_mesh, n_p_mesh = Up_mesh(x_mesh, x_l_tau, x_u_tau, connected_elements, n_connected_elements, n_p_mesh_0, dims_tau)
                except:
                    pass
            # Compute regret
            rt.append(Regret(f_new, np.array(x_new_tau), n_jobs_, model, self.engine))
            # Update iteration parameters
            q_inc += delta_q
            q = int(round(q_inc/5.0)*5.0)
            # Get best values
            x_best, f_best = Best_values(x, f, y, self.sense)
            # Update parameters of AF
            af_params["xi"] *= af_params["xi_decay"]
            af_params["f_best"] = f_best
            # Print results if verbose
            if self.verbose == 1:
                x_print, f_print = Print_results(x_best, f_best, cat_val)
                print(ite, ' ', f_print, *x_print)

        res = Create_results(x_best, f_best, x, f, x_l, x_u, dims, self.max_iter, n_p_design_, self.design, af_params, self.constraints_method, rt, models_constraints, model)

        return res


# *******************************************************
# ****** Import libraries ******
# *******************************************************

import numpy as np
from ..utils.Preprocessing_data import Train_reducer, Train_inverter, Reduce, Inverse

# *******************************************************
# ****** Up_search_space_params ******
# *******************************************************

def Up_search_space_params(x, x_tau, connected_elements, n_elements, d_nc, dims, problem_type, n_components, reducer, inverter, inverter_transform):

    if reducer == "no":
        x_tau_ = x_tau
        connected_elements_ = connected_elements
        reducer_trained = None
        inverter_ = None
    else:
        x_tau_, reducer_trained = Train_reducer(x, d_nc, dims, problem_type, n_components, reducer) 
        if inverter_transform == "no":
            inverter_ = Train_inverter(x, x_tau, dims, inverter)
        elif inverter_transform == "yes":
            inverter_ = reducer_trained
        else:
            pass
        connected_elements_inv = [Inverse(connected_elements[i], inverter_transform, inverter) for i in range(n_elements)]
        connected_elements_ = [Reduce(connected_elements_inv[i], d_nc, dims, problem_type, reducer_trained) for i in range(n_elements)]

    return x_tau_, connected_elements_, reducer_trained, inverter_

# *******************************************************
# ****** Up_mesh ******
# *******************************************************

def Up_mesh(x_mesh, x_l, x_u, connected_elements, n_elements, n_p_mesh, dims):

    def New_boundaries(mesh, n_mesh, dims):

        x_l_new, x_u_new = [], []

        for i in range(n_mesh):
            x_l_new.append([np.min(mesh[i][:,j]) for j in range(dims)])
            x_u_new.append([np.max(mesh[i][:,j]) for j in range(dims)])
        
        return x_l_new, x_u_new
    
    def Detect_equal_limits(x_l, x_u, x_l_new, x_u_new, n_elements, n_p_mesh, dims):

        for i in range(n_elements):
            for j in range(dims):
                if x_l_new[i][j] == x_u_new[i][j]:
                    diff = [(x_u[j] - x_l[j])/(n_p_mesh-1)]
                    x_l_new_temp, x_u_new_temp = np.array([(x_l_new[i][j] - diff)[0], (x_l_new[i][j] + diff)[0]])
                    if x_l_new_temp < x_l[j]:
                        x_l_new[i][j] = x_l[j]
                    else:
                        x_l_new[i][j] = x_l_new_temp
                    if x_u_new_temp > x_u[j]:
                        x_u_new[i][j] = x_l[j]
                    else:
                        x_u_new[i][j] = x_u_new_temp

        return x_l_new, x_u_new
    
    def Find_intersection(x_mesh, x_l_new, x_u_new, n_p_mesh, n_elements, dims):

        def Compute_irr_grids(x_l_new, x_u_new, pts, dims):

            lst = []
            for i in range(dims):
                lst.append(np.linspace(x_l_new[i], x_u_new[i], int(np.round(pts[i]+1))))
            x_mesh_grid = np.meshgrid(*lst)
            
            return np.array(x_mesh_grid).T.reshape(-1, dims)

        def Intersect(A, B):

            _, ncols = A.shape
            dtype={'names':['f{}'.format(i) for i in range(ncols)],
                'formats':ncols * [A.dtype]}
            
            return np.intersect1d(A.view(dtype), B.view(dtype))

        points_mesh_new = [(np.array(x_l_new[i]) - np.array(x_u_new[i]))/(x_mesh[0][0] - x_mesh[0][n_p_mesh+1]) for i in range(n_elements)]
        points_mesh_new = np.abs(points_mesh_new)
        A = Compute_irr_grids(x_l_new[0], x_u_new[0], points_mesh_new[0], dims)
        x_mesh_new_ = []
        x_mesh_new_.append(A)

        for i in range(1, n_elements):
            B = Compute_irr_grids(x_l_new[i], x_u_new[i], points_mesh_new[i], dims)
            intersect_ = Intersect(A, B)
            if not intersect_.tolist():
                x_mesh_new_.append(B)
            else:
                x_mesh_new_[0] = np.vstack((np.concatenate(*[x_mesh_new_]), B))

        if type(x_mesh_new_) is not list:
            x_mesh_new_ = [x_mesh_new_]
        
        n_mesh_new = len(x_mesh_new_)

        return x_mesh_new_, n_mesh_new
        
    def Points_mesh(x_l, x_u, points_mesh, n_mesh, dims):

        l = [[(x_u[i][j] - x_l[i][j]) for i in range(n_mesh)] for j in range(dims)]
        a = [np.prod(l[0][i]) for i in range(n_mesh)]
        pts = a/np.max(a)
        points_mesh_min = int(np.sqrt(points_mesh))

        return [points_mesh_min if int(x*points_mesh) < points_mesh_min else int(x*points_mesh) for x in pts]

    x_l_new, x_u_new = New_boundaries(connected_elements, n_elements, dims)
    x_l_new, x_u_new = Detect_equal_limits(x_l, x_u, x_l_new, x_u_new, n_elements, n_p_mesh, dims)
    pts_mesh = int(np.round(x_mesh[0].shape[0]**(1/dims)))
    n_mesh = int(x_mesh[0].shape[0]/(pts_mesh-1))
    connected_elements_new, n_elements_new = Find_intersection(x_mesh, x_l_new, x_u_new, n_mesh, n_elements, dims)
    x_l_new, x_u_new = New_boundaries(connected_elements_new, n_elements_new, dims)
    points_mesh_new = Points_mesh(x_l_new, x_u_new, n_p_mesh, n_elements_new, dims)
    x_mesh_new = []

    for i in range(n_elements_new):
        lists = [np.linspace(x_l_new[i][j], x_u_new[i][j], points_mesh_new[i]) for j in range(dims)]
        x_mesh_grid = np.meshgrid(*lists)
        x_mesh_new.append(np.array(x_mesh_grid).T.reshape(-1, dims))

    return x_mesh_new, points_mesh_new


# *******************************************************
# ****** Import libraries ******
# *******************************************************

import time
import warnings
import numpy as np
import sympy as sp
import multiprocess as mp
from scipy.stats import qmc
from scipy.stats import chi2
from gpflow.kernels import RBF as RBF_gpflow
from sklearn.gaussian_process.kernels import RBF as RBF_sklearn
from ..utils.Preprocessing_data import U_scaling, U_inverse_scaling, Train_reducer, Train_inverter, Find_reducer, Find_inverter, Reduce
from ..utils.Models import Kernel_discovery

try:
    from GPy.kern import RBF as RBF_gpy
except:
    pass

# *******************************************************
# ****** Space ******
# *******************************************************

def Space(domain):

    """ 
    Gets tuple of dimension of the problem, lower and upper bounds of the continuous variables, values for the discrete variables, and names of the variables.
    """

    dims = len(domain)
    d_nc = 0
    d_ni = 0
    d_nq = 0
    x_l = []
    x_u = []
    int_val = []
    cat_val = []
    names = []

    for i in range(dims):
        try:
            names.append(domain[i]['name'])
        except:
            names = 0
        if domain[i]['type'] == "continuous":
            d_nc += 1
            x_l.append(domain[i]['domain'][0])
            x_u.append(domain[i]['domain'][1])
        elif domain[i]['type'] == "integer":
            d_ni += 1
            int_val.append(domain[i]['domain'])
        elif domain[i]['type'] == "categorical":
            d_nq += 1
            features = domain[i]['domain']
            cat_val.append(features)
        else:
            pass
            
    x_l, x_u = np.array(x_l), np.array(x_u)
    d_nd = d_ni + d_nq
    if d_nd == 0:
        int_val = 0
    if d_nq == 0:
        cat_val = 0
    if names == 0:
        names = None

    return dims, d_nc, d_ni, d_nq, d_nd, x_l, x_u, int_val, cat_val, names

# *******************************************************
# ****** Problem_type ******
# *******************************************************

def Problem_type(d_nc, d_ni, d_nq):

    """ 
    Returns the type of problem depending of the characteristics of the inputs:
    * Continuous if there are only continuous variables
    * Discrete if there are only discrete variables
    * Mixed if there are continuous and discrete variables
    """
    
    if d_nc > 0 and d_ni == 0 and d_nq == 0:
        problem_type = "Continuous"
    elif d_nc == 0 and d_ni > 0 and d_nq == 0:
        problem_type = "Discrete"
    elif d_nc == 0 and d_ni == 0 and d_nq > 0:
        problem_type = "Categorical"
    elif d_nc > 0 and d_ni > 0 and d_nq == 0:
        problem_type = "Mixed_integer"
    elif d_nc > 0 and d_ni == 0 and d_nq > 0:
        problem_type = "Mixed_categorical"
    elif d_nc == 0 and d_ni > 0 and d_nq > 0:
        problem_type = "Mixed_discrete"
    elif d_nc > 0 and d_ni > 0 and d_nq > 0:
        problem_type = "Mixed_all"
    else:
        pass

    return problem_type

# *******************************************************
# ****** Get_constraints ******
# *******************************************************

def Get_constraints(constraints, constraints_method):

    "Transforms tuple of constraints into lists"

    if constraints is None:
        const, d_nqonst = None, None
    else:
        d_nqonst = len(constraints)
        if constraints_method == "PoF":
            symbols = ['<=', '>=']
            const = []
            for i in range(d_nqonst):
                for s in symbols:
                    try:
                        index_const = constraints[i]['constraint'].index(s)
                        const.append(constraints[i]['constraint'][:index_const-1])
                    except:
                        pass
        elif constraints_method == "GPC":
            const = constraints
    
    return const, d_nqonst

# *******************************************************
# ****** Get_p_design ******
# *******************************************************

def Get_n_p_design(fun, d_nc, d_ni, d_nq, x_l, x_u, int_val, cat_val, c1_param, problem_type, design, n_p_design):

    """
    Computes the number of points of the initial design
    """

    # *************************
        # x0_trial
    # *************************

    def x0_trial(x_l, x_u, int_val, cat_val, d_nc, d_ni, d_nq, problem_type):
    
        """ 
        Perform a single random sampling of inputs
        """    

        # x_c0_rand
        def x_c0_rand(x_l, x_u, d_nc):

            return np.random.uniform(x_l, x_u, size=(d_nc))
        
        # x_c0_rand
        def x_d0_rand(disc_val, d_ni):

            return np.array([np.random.choice(disc_val[i]) for i in range(d_ni)])
        
        # Main program
        if problem_type == "Continuous":
            var = x_c0_rand(x_l, x_u, d_nc)
        elif problem_type == "Discrete":
            var = x_d0_rand(int_val, d_ni)
        elif problem_type == "Categorical":
            var = x_d0_rand(cat_val, d_nq)
        elif problem_type == "Mixed_integer":
            xc_var = x_c0_rand(x_l, x_u, d_nc)
            xi_var = x_d0_rand(int_val, d_ni)
            var = np.hstack((xc_var, xi_var))
        elif problem_type == "Mixed_categorical":
            xc_var = x_c0_rand(x_l, x_u, d_nc)
            xq_var = x_d0_rand(cat_val, d_nq)
            var = np.hstack((xc_var, xq_var.astype(object)))
        elif problem_type == "Mixed_discrete":
            xi_var = x_d0_rand(int_val, d_ni)
            xq_var = x_d0_rand(cat_val, d_nq)
            var = np.hstack((xi_var, xq_var.astype(object)))
        elif problem_type == "Mixed_all":
            xc_var = x_c0_rand(x_l, x_u, d_nc)
            xi_var = x_d0_rand(int_val, d_ni)
            xq_var = x_d0_rand(cat_val, d_nq)
            var = np.hstack((xc_var, xi_var, xq_var.astype(object)))
        else:
            pass

        return var

    # *************************
        # Times_fun
    # *************************

    def Times_fun(fun, x):

        """ 
        Determines the computing time for a random evaluation (x0_trial) of the objective function
        """
        
        start = time.time()
        f_eval = fun(x.reshape(1,-1))
        end = time.time()
        time_eval = (end - start)

        if time_eval < 1e-10:
            time_eval = 1e-8
        
        return time_eval, f_eval

    # *************************
        # Points_initial_design
    # *************************

    def Points_initial_design(times, design, c1_param):

        """
        Generates the number of points depending on the cost of the evaluation of the function
        """

        if times <= 1:
            exp_param = 0.25
        else: 
            exp_param = 0.95
        points_D = int(c1_param - c1_param/(1+(1/times)**exp_param))
        # Adjust points if design is Sobol. 
        if design == "Sobol":
            points_D = int(np.ceil(np.sqrt(points_D))**2)
        else:
            pass
        # Adjust points if number is below 3.
        if points_D < 3:
            points_D = int(3)

        return points_D

    # *************************
        # Main program
    # *************************

    x_trial = x0_trial(x_l, x_u, int_val, cat_val, d_nc, d_ni, d_nq, problem_type)
    times, f_trial = Times_fun(fun, x_trial)

    if n_p_design == None:
        n_p_design = Points_initial_design(times, design, c1_param)
    else:
        x_trial, f_trial = None, None

    return n_p_design, x_trial, f_trial

# *******************************************************
# ****** U_Generator ******
# *******************************************************

def U_Generator(dims, points, design_type):

    """ 
    Computes unitary design matrix
    """

    # *************************
    # Random_design
    # *************************

    def Random_design(dims, points):

        return np.random.rand(points, dims)
    
    # *************************
    # QMC_design
    # *************************

    def QMC_design(dims, points, design_type):
        
        # Main program
        if design_type == "LHS":
            method = qmc.LatinHypercube(d=dims)
        elif design_type == "Sobol":
            method = qmc.Sobol(d=dims)
        elif design_type == "Halton":
            method = qmc.Halton(d=dims)
        else:
            pass

        return method.random(n=points)

    # *************************
    # Main program
    # *************************
    
    if design_type == "random":
        variables = Random_design(dims, points)
    elif design_type == "LHS" or design_type == "Sobol" or design_type == "Halton":
        variables = QMC_design(dims, points, design_type)
    else:
        pass
    
    return variables

# *******************************************************
# ****** Get_x_and_z ******
# *******************************************************

def Get_x_and_z(fun, x_0, f_0, dims, d_nc, d_ni, d_nq, x_l, x_u, int_val, cat_val, c1_param, problem_type, design_type, n_p_design):

    """ 
    Gets the training data in the problem bounds
    """

    if x_0 is None:
        if n_p_design is None:
            n_p_design, x_trial, f_trial = Get_n_p_design(fun, d_nc, d_ni, d_nq, x_l, x_u, int_val, cat_val, c1_param, problem_type, design_type, n_p_design)
            U = U_Generator(dims, n_p_design, design_type)
            x_0 = U_scaling(U, d_nc, d_ni, d_nq, x_l, x_u, int_val, cat_val, n_p_design, problem_type)
            f_0 = fun(x_0).reshape(-1,1)
            x, f = np.vstack((x_0, x_trial)), np.vstack((f_0, f_trial))
        else:
            U = U_Generator(dims, n_p_design, design_type)
            x = U_scaling(U, d_nc, d_ni, d_nq, x_l, x_u, int_val, cat_val, n_p_design, problem_type)
            f = fun(x).reshape(-1,1)
    else: 
        x = x_0
        if f_0 is None:
            f = fun(x).reshape(-1,1)
        else:
            f = f_0
        n_p_design = len(f)

    return x, f, n_p_design

# *******************************************************
# ****** Get_constraints ******
# *******************************************************

def Get_constraints(constraints, constraints_method):

    """ 
    Transforms tuple of constraints into list
    """

    if constraints is None:
        const, n_const = None, None
    else:
        n_const = len(constraints)
        if constraints_method == "PoF":
            symbols = ['<=', '>=']
            const = []
            for i in range(n_const):
                for s in symbols:
                    try:
                        if s == '<=':
                            index_const = constraints[i]['constraint'].index(s)
                            const.append(constraints[i]['constraint'][:index_const-1])
                        elif s == '>=':
                            index_const = constraints[i]['constraint'].index(s)
                            const.append(-constraints[i]['constraint'][:index_const-1])
                    except:
                        pass
        elif constraints_method == "GPC":
            const = constraints
    
    return const, n_const

# *******************************************************
# ****** Get_search_space_params ******
# *******************************************************

def Get_search_space_params(x, dims, d_nc, d_ni, d_nq, d_nd, x_l, x_u, int_val, cat_val, points, problem_type, reducer, inverter_transform):

    """ 
    Return the parameters for the search space of the new points:
    * If dimension reduction is performed, it returns the training set in the reduced space and the trained reduction method. 
    """

    if reducer == "no":
        if dims > 6: warnings.warn('You can reach maximum depth of recursion')
        if (problem_type == "Continuous") and (d_nc <= 5):
            reducer_name, reducer_trained, dims_tau, x_tau = "no", None, dims, x
        else:
            reducer_name, reducer_trained, dims_tau = "no", None, dims
            x_tau = U_inverse_scaling(x, d_nc, d_ni, d_nq, x_l, x_u, int_val, cat_val, points, problem_type)
    elif reducer is None:
        if (problem_type == "Continuous") and (d_nc <= 5):
            reducer_name, reducer_trained, dims_tau, x_tau = "no", None, dims, x
        else:
            reducer_name, reducer_trained, dims_tau = Find_reducer(x, d_nc, dims, problem_type)
            x_tau = Reduce(x, d_nc, dims, problem_type, reducer_trained)
    else:
        dims_tau = reducer.n_components
        reducer_name = reducer
        x_tau, reducer_trained = Train_reducer(x, d_nc, dims, problem_type, dims_tau, reducer)

    if inverter_transform == "yes":
        if reducer_name.__module__ == 'prince.mca' or reducer_name.__module__ == 'prince.famd':
            raise ValueError("Reducer module has no inverse_transform")
        else:
            inverter = reducer_trained
    elif inverter_transform == "no":
        if reducer_name == "no":
            inverter = None
        else:
            inverter = Find_inverter(x, x_tau, d_nc, d_nd, dims, problem_type)
            inverter = Train_inverter(x, x_tau, dims, inverter)
    else:
        pass
    
    return x_tau, reducer_name, reducer_trained, inverter, dims_tau

# *******************************************************
# ****** Get_x_mesh ******
# *******************************************************

def Get_x_mesh(lower_bound, upper_bound, dims_tau, c2_param):

    """ 
    
    """
    # *************************
    # Random_design
    # *************************

    def Points_mesh(dims, c2_param):

        points = int(np.ceil(2**(c2_param/dims)))
        if points < 3:
            points = 3

        return points
    
    # *************************
    # Mesh_design
    # *************************

    def Mesh_design(lower_bound, upper_bound, dims, points):

        lists = [np.linspace(lower_bound[i], upper_bound[i], points) for i in range(dims)]
        mesh = np.meshgrid(*lists)
        
        return np.array(mesh).T.reshape(-1, dims)
    
    # *************************
    # Main program
    # *************************
    
    n_p_mesh = Points_mesh(dims_tau, c2_param)
    x_mesh = Mesh_design(lower_bound, upper_bound, dims_tau, n_p_mesh)
    
    return [x_mesh], [n_p_mesh], n_p_mesh

# *******************************************************
# ****** Bounds ******
# *******************************************************

def Bounds(x_l, x_u, dims):

    bnds = np.sort(np.array([[x_l[i], x_u[i]] for i in range(dims)]))
    return tuple(map(tuple, bnds))

# *******************************************************
# ****** Get_n_jobs ******
# *******************************************************

def Get_n_jobs(n_jobs):

    if n_jobs == -1:
        jobs = mp.cpu_count()
    elif n_jobs == None:
        jobs = 1
    else:
        jobs = n_jobs
    
    return jobs

# *******************************************************
# ****** Get_kernel ******
# *******************************************************

def Get_kernel(x, z, dims, surrogate, beta, kernel, kern_discovery, kern_discovery_evals, engine):

    if kern_discovery == "yes":
        kernel_ = Kernel_discovery(x, z, dims, surrogate, beta, kern_discovery_evals, engine)
    elif kern_discovery == "no" and kernel is None:
        if engine == 'gpflow':
            kernel_ = RBF_gpflow()
        elif engine == 'sklearn':
            kernel_ = RBF_sklearn()
        elif engine == 'GPy':
            kernel_ = RBF_gpy(input_dim=dims, variance=1.0, lengthscale=1.0)
        else:
            pass
    else:
        kernel_ = kernel

    return kernel_

# *******************************************************
# ****** Iter_params ******
# *******************************************************

def Iter_params(dims, dims_tau, max_iter, alpha):

    x_symb = sp.Matrix(sp.symbols('x:' + str(dims_tau)))
    x_symb_names = sp.Matrix(sp.symbols('x:' + str(dims)))
    flag = 0
    q0 = 25
    qf = 95
    delta_q = (qf-q0)/max_iter
    q_inc = q0
    
    if dims > 1:
        chi = chi2.ppf(alpha, dims_tau)
    else:
        chi = None

    if max_iter < 10:
        update_param = 1
    elif max_iter < 51:
        update_param = int(max_iter/10)
    elif max_iter < 101:
        update_param = 5
    elif max_iter < 501:
        update_param = 10
    elif max_iter < 1001:
        update_param = 50
    else:
        pass

    return x_symb, x_symb_names, flag, q0, q_inc, delta_q, chi, update_param

# *******************************************************
# ****** AF_params ******
# *******************************************************

def AF_params(z, xi_0, xi_f, xi_decay, iters, AF_name, sense):

    if xi_decay == "yes":
        xi_decay = (xi_f/xi_0)**(1/iters)
    elif xi_decay == "no":
        xi_decay = 1
    else:
        pass

    if sense == "maximize":
        f_best = np.max(z)
    elif sense == "minimize":
        f_best = np.min(z)
    else:
        pass

    return {'xi': xi_0, 'xi_decay': xi_decay, 'f_best': f_best, 'AF_name': AF_name}


import os
from setuptools import find_packages, setup

filename = os.path.abspath(os.path.dirname(__file__))

with open(os.path.join(filename, 'README.md'), encoding='utf-8') as f:
    README = f.read()

setup(
    name='pyBOWIE',
    version='1.0.3',
    description='Bayesian optimization for constrained or unconstrained, continuous, discrete or mixed data problems',
    author='Javier Morlet-Espinosa, Antonio Flores-Tlacuahuac',
    author_email='a00833961@tec.mx',
    url='https://github.com/JavierMorlet/pyBOWIE',
    long_description=README,
    long_description_content_type="text/markdown",
    packages = find_packages(),
    install_requires=[
        'numpy>=1.23.5',
        'sympy>=1.11.1',
        'pandas>=2.0.3',
        'GPy>=1.10.0',
        'gpflow >= 2.9.2',
        'scipy>=1.10.1',
        'scikit-learn>=1.1.3',
        'properscoring>=0.1',
        'prince>=0.12.1',
        'multiprocess >= 0.70.16',
        'matplotlib>=3.7.3'
        ]
    )


# *******************************************************
# ****** Import libraries ******
# *******************************************************

import numpy as np
import pandas as pd
from prince import PCA, MCA, FAMD
from sklearn.metrics import r2_score
from sklearn.ensemble import ExtraTreesRegressor, ExtraTreesClassifier
from sklearn.model_selection import RandomizedSearchCV, train_test_split

# *******************************************************
# ****** U_scaling ******
# *******************************************************

def U_scaling(U, d_nc, d_ni, d_nq, x_l, x_u, int_val, cat_val, points, problem_type):

    def Scale_continuous(U, x_l, x_u):
        
        return (x_u - x_l) * U + x_l

    def Scale_discrete(U, d_nd, disc_val, points):

        size_y = [len(disc_val[i]) for i in range(d_nd)]
        l = [[i/size_y[j] for i in range(size_y[j]+1)] for j in range(d_nd)]
        y_new = np.empty([points, d_nd], dtype=object)
        for i in range(len(l)):
            l[i][-1] = l[i][-1] + 0.0001
        for i in range(d_nd):
            for j in range(len(U[:,i])):
                for k in range(len(l[i])-1):
                    if l[i][k] <= U[j,i] < l[i][k+1]:
                        y_new[j,i] = disc_val[i][k]
                        break
        
        return y_new

    if problem_type == "Continuous":
        vars = Scale_continuous(U, x_l, x_u)
    elif problem_type == "Discrete":
        vars = Scale_discrete(U, d_ni, int_val, points).astype(float)
    elif problem_type == "Categorical":
        vars = Scale_discrete(U, d_nq, cat_val, points)
    elif problem_type == "Mixed_integer":
        xc_vars = Scale_continuous(U[:,:d_nc], x_l, x_u)
        if d_ni == 1:
            xi_vars = Scale_discrete(U[:,d_nc:].reshape(-1,1), d_ni, int_val, points)
        else:
            xi_vars = Scale_discrete(U[:,d_nc:], d_ni, int_val, points)
        vars = np.hstack((xc_vars, xi_vars)).astype(float)
    elif problem_type == "Mixed_categorical":
        xc_vars = Scale_continuous(U[:,:d_nc], x_l, x_u)
        if d_ni == 1:
            xq_vars = Scale_discrete(U[:,d_nc:].reshape(-1,1), d_nq, cat_val, points)
        else:
            xq_vars = Scale_discrete(U[:,d_nc:], d_nq, cat_val, points)
        vars = np.hstack((xc_vars, xq_vars))
    elif problem_type == "Mixed_discrete":
        if d_ni == 1:
            xi_vars = Scale_discrete(U[:,0:d_ni].reshape(-1,1), d_ni, int_val, points)
        else:
            xi_vars = Scale_discrete(U[:,0:d_ni], d_ni, int_val, points)
        if d_ni == 1:
            xq_vars = Scale_discrete(U[:,d_ni:].reshape(-1,1), d_nq, cat_val, points)
        else:
            xq_vars = Scale_discrete(U[:,d_ni:], d_nq, cat_val, points)
        vars = np.hstack((xi_vars, xq_vars))
    elif problem_type == "Mixed_all":
        xc_vars = Scale_continuous(U[:,:d_nc], x_l, x_u)
        d_nt = d_nc + d_ni
        if d_ni == 1:
            xi_vars = Scale_discrete(U[:,d_nc:d_nt].reshape(-1,1), d_ni, int_val, points)
        else:
            xi_vars = Scale_discrete(U[:,d_nc:d_nt], d_ni, int_val, points)
        if d_ni == 1:
            xq_vars = Scale_discrete(U[:,d_nt:].reshape(-1,1), d_nq, cat_val, points)
        else:
            xq_vars = Scale_discrete(U[:,d_nt:], d_nq, cat_val, points)
        vars = np.hstack((xc_vars, xi_vars, xq_vars))
    else:
        pass

    return vars

# *******************************************************
# ****** U_scaling ******
# *******************************************************

def U_inverse_scaling(X, d_nc, d_ni, d_nq, x_l, x_u, int_val, cat_val, points, problem_type):

    def Scale_continuous_inv(X, x_l, x_u):
        
        return (X - x_l)/(x_u - x_l)
    
    def Scale_discrete_inv(X, d_nd, disc_val, points):

        y_new = np.empty([points+1, d_nd])
        size_y = [len(disc_val[i]) for i in range(d_nd)]
        for i in range(d_nd):
            #print(int_val)
            for j in range(len(disc_val[i])):
                y_new[:,i][np.where(disc_val[i][j] == X[:,i])[0].tolist()] = j/size_y[i]
        
        return y_new

    if problem_type == "Continuous":
        vars = Scale_continuous_inv(X, x_l, x_u)
    elif problem_type == "Discrete":
        vars = Scale_discrete_inv(X, d_ni, int_val, points).astype(float)
    elif problem_type == "Categorical":
        vars = Scale_discrete_inv(X, d_nq, cat_val, points)
    elif problem_type == "Mixed_integer":
        xc_vars = Scale_continuous_inv(X[:,:d_nc], x_l, x_u)
        if d_ni == 1:
            xi_vars = Scale_discrete_inv(X[:,d_nc:].reshape(-1,1), d_ni, int_val, points)
        else:
            xi_vars = Scale_discrete_inv(X[:,d_nc:], d_ni, int_val, points)
        vars = np.hstack((xc_vars, xi_vars)).astype(float)
    elif problem_type == "Mixed_categorical":
        xc_vars = Scale_continuous_inv(X[:,:d_nc], x_l, x_u)
        if d_nq == 1:
            xq_vars = Scale_discrete_inv(X[:,d_nc:].reshape(-1,1), d_nq, cat_val, points)
        else:
            xq_vars = Scale_discrete_inv(X[:,d_nc:], d_nq, cat_val, points)
        vars = np.hstack((xc_vars, xq_vars))
    elif problem_type == "Mixed_discrete":
        if d_ni == 1:
            xi_vars = Scale_discrete_inv(X[:,0:d_ni].reshape(-1,1), d_ni, int_val, points)
        else:
            xi_vars = Scale_discrete_inv(X[:,0:d_ni], d_ni, int_val, points)
        if d_ni == 1:
            xq_vars = Scale_discrete_inv(X[:,d_ni:].reshape(-1,1), d_nq, cat_val, points)
        else:
            xq_vars = Scale_discrete_inv(X[:,d_ni:], d_nq, cat_val, points)
        vars = np.hstack((xi_vars, xq_vars))
    elif problem_type == "Mixed_all":
        xc_vars = Scale_continuous_inv(X[:,:d_nc], x_l, x_u)
        d_nt = d_nc + d_ni
        if d_ni == 1:
            xi_vars = Scale_discrete_inv(X[:,d_nc:d_nt].reshape(-1,1), d_ni, int_val, points)
        else:
            xi_vars = Scale_discrete_inv(X[:,d_nc:d_nt], d_ni, int_val, points)
        if d_ni == 1:
            xq_vars = Scale_discrete_inv(X[:,d_nt:].reshape(-1,1), d_nq, cat_val, points)
        else:
            xq_vars = Scale_discrete_inv(X[:,d_nt:], d_nq, cat_val, points)
        vars = np.hstack((xc_vars, xi_vars, xq_vars))
    else:
        pass

    return vars

# *******************************************************
# ****** Trans_data_to_pandas ******
# *******************************************************

def Trans_data_to_pandas(data, d_nc, dims, problem_type):

    if problem_type == "Mixed_categorical" or problem_type == "Mixed_discrete" or problem_type == "Mixed_all":
        cols = np.arange(1,dims+1)
        cols = cols.astype(str)
        data = pd.DataFrame(data, columns=cols)
        data[cols[0:d_nc]] = data[cols[0:d_nc]].astype(float)
    else:
        data = pd.DataFrame(data)

    return data

# *******************************************************
# ****** Reduce ******
# *******************************************************

def Reduce(data, d_nc, dims, problem_type, reducer):
    
    data_pd = Trans_data_to_pandas(data, d_nc, dims, problem_type)
    
    return np.array(reducer.transform(data_pd))

# *******************************************************
# ****** Inverse ******
# *******************************************************

def Inverse(data, inverter_transform, inverter):

    def Back_projection(data, model):

        predictions = {}
        for col, mod in model.items():
            predictions[col] = mod.predict(data)
        
        return predictions
    
    data = pd.DataFrame(data)
    if inverter_transform == "yes":
        data_reconstructed = inverter.inverse_transform(data)
    elif inverter_transform == "no":
        predictions = Back_projection(data, inverter)
        data_reconstructed = np.vstack(([predictions[i] for i in range(len(inverter))])).T

    return data_reconstructed

# *******************************************************
# ****** Train_reducer ******
# *******************************************************

def Train_reducer(data, d_nc, dims, problem_type, n_components, reducer):

    data_pd = Trans_data_to_pandas(data, d_nc, dims, problem_type)
    if reducer.__module__ == 'prince.pca':
        reducer_train = PCA()
    elif reducer.__module__ == 'prince.famd':
        reducer_train = FAMD()
    elif reducer.__module__ == 'prince.mca':
        reducer_train = MCA()
    else:
        reducer_train = reducer

    reducer_train.n_components = n_components
    data_red = np.array(reducer_train.fit_transform(data_pd))
    
    return data_red, reducer_train

# *******************************************************
# ****** Train_inverter ******
# *******************************************************

def Train_inverter(data, data_reduced, dims, models):

    for i in range(dims):
        model = models[i]
        n_estimators_ = model.n_estimators
        model.set_params(n_estimators=n_estimators_+5)
        model.fit(data_reduced, data[:,i])
        models[i] = model

    return models

# *******************************************************
# ****** Find_reducer ******
# *******************************************************

def Find_reducer(data, d_nc, dims, problem_type):

    def Find_n_components_PCA(x_train, x_test, n, reducer, error_min = 0.25, n_max = 5):

        def Train_Red_comp(x_train, n_trial, reducer):

            reducer.n_components = n_trial
            reducer = reducer.fit(x_train)
            
            return reducer
        
        def Reconstruction_error(x_test, reducer):

            X_test_red = reducer.transform(x_test)
            X_test_reconstructed = reducer.inverse_transform(X_test_red)
            
            return r2_score(x_test, X_test_reconstructed)

        for n in range(1, n_max):
            reducer = Train_Red_comp(x_train, n, reducer)
            error = Reconstruction_error(x_test, reducer)
            if error >= error_min:
                break

        return n
    
    def Find_n_components(data, n, reducer, var_max = 75, n_max = 6):

        def Red(reducer, data, n_trial):

            reducer.n_components = n_trial
            reducer = reducer.fit(data)

            return reducer.cumulative_percentage_of_variance_[-1]
        
        for n in range(1, n_max):
            variance = Red(reducer, data, n)
            if variance >= var_max:
                break

        return n

    # Main program
    
    if problem_type == "Continuous" or problem_type == "Discrete" or problem_type == "Mixed_integer":
        reducer = PCA()
        reducer_search = PCA()
        reducer_train = PCA()
    elif problem_type == "Mixed_categorical" or problem_type == "Mixed_discrete" or problem_type == "Mixed_all":
        reducer = FAMD()
        reducer_train = FAMD()
    elif problem_type == "Categorical":
        reducer = MCA()
        reducer_train = MCA()
    else:
        pass

    data_pd = Trans_data_to_pandas(data, d_nc, dims, problem_type)
    if problem_type == "Continuous":
        x_train, x_test = train_test_split(data_pd)
        dims_red = Find_n_components_PCA(x_train, x_test, dims, reducer_search)
        reducer_train = PCA(n_components=dims_red)
        reducer_train = reducer_train.fit(data_pd)
    else:
        dims_red = Find_n_components(data_pd, dims, reducer_train)

    return reducer, reducer_train, dims_red

# *******************************************************
# ****** Find_inverter ******
# *******************************************************

def Find_inverter(data, data_reduced, d_nc, d_nd, dims, problem_type):

    def Find_best_models(x, z, dims, problem_type, method):

        param_clas = {
            'n_estimators': [50, 100, 200],
            'max_depth': [None, 10, 20, 30],
            'min_samples_split': [2, 5, 10],
            'min_samples_leaf': [1, 2, 4],
            'bootstrap': [True, False]
        }

        if (problem_type == "Discrete" and dims < 5) or (problem_type == "Categorical" and dims < 5) or (problem_type == "Mixed_discrete" and dims < 5):
            random_search = RandomizedSearchCV(method(), param_distributions=param_clas, n_iter=15, cv=dims)
        elif (problem_type == "Mixed_integer" and d_nd < 5) or (problem_type == "Mixed_categorical" and d_nd < 5) or (problem_type == "Mixed_all" and d_nd < 5):
            random_search = RandomizedSearchCV(method(), param_distributions=param_clas, n_iter=15, cv=d_nd)
        else:
            random_search = RandomizedSearchCV(method(), param_distributions=param_clas, n_iter=15, cv=5)
        random_search.fit(x, z)
        best_params = random_search.best_params_
        best_model = method(**best_params)
        
        return best_model

    def Train_models(x, z, problem_type, method):

        models = {}
        for col in z.columns:
            model = Find_best_models(x, z[col], problem_type, dims, method)
            model.warm_start = True
            models[col] = model

        return models
    
    data_pd = Trans_data_to_pandas(data, d_nc, dims, problem_type)

    if problem_type == "Continuous":
        model = Train_models(data_reduced, data_pd, problem_type, ExtraTreesRegressor)
    elif problem_type == "Discrete" or problem_type == "Categorical" or problem_type == "Mixed_discrete":
        model = Train_models(data_reduced, data_pd, problem_type, ExtraTreesClassifier)
    elif problem_type == "Mixed_integer" or problem_type == "Mixed_categorical" or problem_type == "Mixed_all":
        model_1 = Train_models(data_reduced, data_pd.iloc[:,:d_nc], problem_type, ExtraTreesRegressor)
        model_2 = Train_models(data_reduced, data_pd.iloc[:,d_nc:], problem_type, ExtraTreesClassifier)
        model = {**model_1, **model_2}
        model_names = np.arange(len(model))
        model = dict(zip(model_names, list(model.values())))

    return model

# *******************************************************
# ****** Red_bounds ******
# *******************************************************

def Red_bounds(x_l, x_u, int_val, cat_val, d_nc, d_ni, d_nq, dims, dims_tau, problem_type, reducer):

    if reducer is None:
        if problem_type == "Continuous":
            x_l_tau, x_u_tau = x_l, x_u
        else:
            x_l_tau, x_u_tau = np.zeros(dims), np.ones(dims)
    else:
        if problem_type == "Continuous":
            bounds = [[x_l[i], x_u[i]] for i in range(dims)]
        elif problem_type == "Discrete":
            bounds = [[int_val[i][0], int_val[i][-1]] for i in range(dims)]
        elif problem_type == "Mixed_integer":
            xc_bounds = [[x_l[i], x_u[i]] for i in range(d_nc)]
            xi_bounds = [[int_val[i][0], int_val[i][-1]] for i in range(d_ni)]
            bounds = np.vstack((xc_bounds, xi_bounds)).tolist()
        elif problem_type == "Categorical":
            bounds = [[cat_val[i][j] for j in range(len(cat_val[i]))] for i in range(dims)]
        elif problem_type == "Mixed_categorical":
            xc_bounds = [[x_l[i], x_u[i]] for i in range(d_nc)]
            xq_bounds = [[cat_val[i][j] for j in range(len(cat_val[i]))] for i in range(d_nq)]
            bounds = np.vstack((xc_bounds, xq_bounds)).tolist()
        elif problem_type == "Mixed_discrete":
            xi_bounds = [[int_val[i][0], int_val[i][-1]] for i in range(d_ni)]
            xq_bounds = [[cat_val[i][j] for j in range(len(cat_val[i]))] for i in range(d_nq)]
            bounds = np.vstack((xi_bounds, xq_bounds)).tolist()
        elif problem_type == "Mixed_discrete":
            xc_bounds = [[x_l[i], x_u[i]] for i in range(d_nc)]
            xi_bounds = [[int_val[i][0], int_val[i][-1]] for i in range(d_ni)]
            xq_bounds = [[cat_val[i][j] for j in range(len(cat_val[i]))] for i in range(d_nq)]
            bounds = np.vstack((xc_bounds, xi_bounds, xq_bounds)).tolist()
        mesh = np.meshgrid(*bounds)
        x_mesh = np.array(mesh).T.reshape(-1, dims)
        x_mesh_red = Reduce(x_mesh, d_nc, dims, problem_type, reducer)
        x_l_tau = np.array([x_mesh_red[0,i] for i in range(dims_tau)])
        x_u_tau = np.array([x_mesh_red[-1,i] for i in range(dims_tau)])

    return x_l_tau, x_u_tau

# *******************************************************
# ****** Red_bounds ******
# *******************************************************

def X_new_scaling(x_new_tau, d_nc, d_ni, d_nq, x_l, x_u, int_val, cat_val, jobs, reducer, inverter_transform, inverter, problem_type):

    if reducer == "no":
        if problem_type == "Continuous":
            x_new = np.array(x_new_tau)
        else:
            x_new_tau = np.array(x_new_tau)
            x_new = U_scaling(x_new_tau, d_nc, d_ni, d_nq, x_l, x_u, int_val, cat_val, jobs, problem_type)
    else:
        x_new = Inverse(x_new_tau, inverter_transform, inverter)
        x_new = np.array(x_new)

    return x_new